from subprocess import check_call, CalledProcessError

import logging


def require_clean_work_tree() -> bool:
    check_call(["git", "update-index", "-q", "--ignore-submodules", "--refresh"])

    try:
        check_call(["git", "diff-files", "--quiet", "--ignore-submodules", "--"])
    except CalledProcessError:
        logging.error("You have unstaged changes.")
        return False

    try:
        check_call(["git", "diff-index", "--cached", "--quiet", "HEAD", "--ignore-submodules", "--"])
    except CalledProcessError:
        logging.error("Your index contains uncommitted changes.")
        return False

    return True



import logging
from os.path import dirname, join, isdir, isfile, basename

import docker
from docker.types import LogConfig
from paramiko import SSHClient, SSHException
from paramiko.ssh_exception import NoValidConnectionsError

from yog.host.manage_docker_utils import is_acceptable_container, my_format_to_run_ports_arg_format, build_volumes_dict
from yog.ssh_utils import check_call, check_stdout, ScopedProxiedRemoteSSHTunnel, compare_local_and_remote
from yog.host import necronomicon
from yog.host.necronomicon import Necronomicon

log = logging.getLogger(__name__)


def apply_necronomicon(host: str, root_dir):
    ssh = SSHClient()
    ssh.load_system_host_keys()
    try:
        log.info(f"[{host}]")
        ssh.connect(host)
        apply_necronomicon_for_host(host, ssh, root_dir)
    except RuntimeError as e:
        log.error(f"{host} error: {e.__class__.__name__}: {str(e)}")
    except SSHException as e:
        log.error(f"{host} error: {e.__class__.__name__}: {str(e)}")
    except NoValidConnectionsError as e:
        log.error(f"{host} error: {e.__class__.__name__}: {str(e)}")
    finally:
        ssh.close()


def load_necronomicons_for_host(host: str, root_dir):
    necronomicon_paths = []
    cur = join(root_dir, "domains")
    if isfile(join(cur, "_.yml")):
        necronomicon_paths.append(join(cur, "_.yml"))

    for part in reversed(host.split(".")):
        if not isdir(cur):
            break
        if isfile(join(cur, f"{part}.yml")):
            necronomicon_paths.append(join(cur, f"{part}.yml"))
        cur = join(cur, part)

    return [necronomicon.loadfile(basename(p), p) for p in necronomicon_paths]


def apply_necronomicon_for_host(host: str, ssh: SSHClient, root_dir):
    necronomicons = load_necronomicons_for_host(host, root_dir)
    if not necronomicons:
        raise RuntimeError(f"No necronomicons found for {host}")

    for n in necronomicons:
        if n.files.files:
            apply_files_section(host, n, ssh, root_dir)
        if n.docker.containers:
            apply_docker_section(host, n, ssh)
        if n.cron.crons:
            apply_cron_section(host, n, ssh)


def apply_cron_section(host: str, n: Necronomicon, ssh: SSHClient):
    cronfile_lines = []
    line1_length = max([len(c.expr) for c in n.cron.crons])
    line2_length = max([len(c.user) for c in n.cron.crons])
    for cron in n.cron.crons:
        cronfile_lines.append(f"{cron.expr.ljust(line1_length, ' ')}\t{cron.user.ljust(line2_length, ' ')}\t{cron.command}")
    cronfile_body = "\n".join(cronfile_lines + [""])
    ok, expected, found = compare_local_and_remote(bytes(cronfile_body, "utf-8"), "/etc/cron.d/yog.cron", ssh)
    if not ok:
        log.info(f"[{host}][cron]: stale")
        log.info(f"Writing /etc/cron.d/yog.cron version {expected[:10]}")
        check_call(ssh, "sudo bash -c 'cat - > /etc/cron.d/yog.cron'", send_stdin=cronfile_body)
    else:
        log.info(f"[{host}][cron]: ok")


def apply_docker_section(host: str, n: Necronomicon, ssh: SSHClient):
    log.debug(f"Docker: {n.ident}")
    tunnels = []
    for tunnel_def in n.tunnels.tunnels:
        log.debug(f"Setting up tunnel {tunnel_def}")
        tunnel = ScopedProxiedRemoteSSHTunnel(
            host,
            tunnel_def.target_port,
            tunnel_def.host,
            "remote",
            force_random_port=tunnel_def.local_port)
        tunnels.append(tunnel)

    try:
        for tun in tunnels:
            tun.connect()
        with ScopedProxiedRemoteSSHTunnel(host, 4243) as tport:
            log.debug(f"Connecting docker client.... tcp://localhost:{tport}")
            client = docker.DockerClient(base_url=f"tcp://localhost:{tport}", version="auto")
            log.debug("Docker connected.")

            for desired_container in n.docker.containers:
                log.debug(desired_container)

                desired_container_env = {}
                for name, value in desired_container.env.items():
                    value = str(value)
                    if value.startswith("yogreadfile:"):
                        try:
                            desired_container_env[name] = "\n".join(check_stdout(ssh, f"sudo cat {value[len('yogreadfile:'):]}")).strip()
                        except RuntimeError as err:
                            log.error(f"Error processing yogreadfile: {value}", exc_info=err)
                            raise RuntimeError(f"Error accessing file: {value[len('yogreadfile:'):]}")
                    else:
                        desired_container_env[name] = value

                log.debug(f"PULL: {desired_container.image}@{desired_container.fingerprint}")
                img = client.images.pull(f"{desired_container.image}@{desired_container.fingerprint}")

                cur = client.containers.list(all=True, filters={"name": desired_container.name})
                matches = []
                for c in cur:
                    log.debug(f"Existing container {c.id} is {c.status}")
                    if is_acceptable_container(c, img, desired_container, desired_container_env):
                        log.debug(f"{c.id} is image {c.image.id} which matches our target.")
                        log.info(f"[{host}][docker]: OK {desired_container.name}@{desired_container.fingerprint}")
                        matches.append(c)
                    else:
                        if c.status in ["running", "restarting"]:
                            log.debug(f"STOP {c.name}:{c.id}")
                            c.stop()
                        log.debug(f"RM: {c.name}:{c.id}")
                        c.remove()
                if len(matches) > 0:
                    log.debug("Existing container matches our desired state. No need to kill it.")
                else:
                    log.debug(f"RUN: {desired_container.image}@{desired_container.fingerprint}")
                    log.info(f"[{host}][docker]: stale {desired_container.name}")
                    ports_dict = my_format_to_run_ports_arg_format(desired_container.ports)
                    volumes_dict = build_volumes_dict(desired_container.volumes)
                    client.containers.run(f"{desired_container.image}@{desired_container.fingerprint}",
                                          name=desired_container.name,
                                          restart_policy={'Name': "always"},
                                          volumes=volumes_dict,
                                          ports=ports_dict,
                                          log_config=LogConfig(type=LogConfig.types.JOURNALD),
                                          environment=desired_container_env,
                                          detach=True,
                                          command=desired_container.command)
    finally:
        for tun in tunnels:
            try:
                tun.disconnect()
            except RuntimeError as e:
                log.warning("Error while disconnecting tunnel", exc_info=e)


def apply_files_section(host: str, n: Necronomicon, ssh: SSHClient, root_dir):
    log.debug(f"Files: {n.ident}")
    hupcmds = set()
    for f in n.files.files:
        with open(join(root_dir, "files", f.src)) as fin:
            content = fin.read()
        ok, _, _ = compare_local_and_remote(content.encode("utf-8"), f.dest, ssh, f.root)
        if ok:
            log.info(f"[{host}][files]: OK [{f.src}]")
        else:
            log.info(f"[{host}][files]: stale {f.src} -> {f.dest}")
            if f.hupcmd:
                hupcmds.add(f.hupcmd)

            if f.root:
                cmd_prefix = "sudo "
            else:
                cmd_prefix = ""

            check_call(ssh, f"{cmd_prefix}mkdir -p \"{dirname(f.dest)}\"")
            check_call(ssh, f"{cmd_prefix}bash -c 'cat - > \"{f.dest}\"'", send_stdin=content)
    for c in hupcmds:
        log.info(f"[{host}][files][hup]: {c}")
        check_call(ssh, c)



import re
import typing as t

from docker.models.containers import Container
from docker.models.images import Image

from yog.host.necronomicon import DockerContainer


def my_format_to_run_ports_arg_format(port_section: t.Mapping[str, str]):
    def format_host_expr(host_expr: str) -> t.Tuple[str, str]:
        md: re.Match = re.fullmatch(r"((?P<addr>\d+\.\d+\.\d+\.\d+):)?(?P<port>\d+)(/(?P<proto>tcp|udp))?",
                                host_expr)  # 0.0.0.0:53/tcp

        return md.group("addr") if md.group("addr") else '0.0.0.0', md.group("port")

    def format_container_port_expr(container_port_expr: str) -> str:
        return container_port_expr if container_port_expr.endswith("/tcp") or container_port_expr.endswith(
            "/udp") else f"{container_port_expr}/tcp"

    ret = dict()
    for k, v in port_section.items():
        new_key = format_container_port_expr(v)
        new_value = format_host_expr(k)

        if new_key in ret:
            ret[new_key].append(new_value)
        else:
            ret[new_key] = [new_value]

    #print(ret)
    return ret


def build_volumes_dict(volumes_section: t.Mapping[str, str]):
    ret = {}
    for k, v in volumes_section.items():
        mode_data = re.search(r"\+(?P<mode>r[ow])$", v)
        if mode_data:
            ret[k] = {"bind": v[:-3], "mode": mode_data.group("mode").strip()}
        else:
            ret[k] = {"bind": v, "mode": "rw"}

    return ret


def my_format_to_ports_attr_format(port_section: t.Mapping[str, str]):
    """Converts my YAML format, which is the same as the docker CLI -p format, to what the docker API expects
    MY format is:
        key: <host addr>:<host port>/<proto>
        value: <dest container port>/<proto>
    What the API expects is:
    A dict where the keys are a str of format: "container port/proto"
           and the values are a list whose elements are
                              a dict that looks like {HostIp: <HostAddr>, HostPort: <HostPort>}
    Example:
        {
            '33200/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '33200'}],
            '53/tcp': [{'HostIp': '192.168.1.103', 'HostPort': '53'}],
            '53/udp': [{'HostIp': '192.168.1.103', 'HostPort': '53'}]
        }
    """

    def format_host_expr(host_expr: str) -> t.List[t.Dict[str, str]]:
        md: re.Match = re.fullmatch(r"((?P<addr>\d+\.\d+\.\d+\.\d+):)?(?P<port>\d+)(/(?P<proto>tcp|udp))?",
                                host_expr)  # 0.0.0.0:53/tcp

        return [{'HostIp': md.group("addr") if md.group("addr") else '0.0.0.0',
                 'HostPort': md.group('port')}]

    def format_container_port_expr(container_port_expr: str) -> str:
        return container_port_expr if container_port_expr.endswith("/tcp") or container_port_expr.endswith(
            "/udp") else f"{container_port_expr}/tcp"

    #ret = {format_container_port_expr(v): format_host_expr(k) for k, v in port_section.items()}
    ret = dict()
    for k, v in port_section.items():
        key = format_container_port_expr(v)
        value = format_host_expr(k)

        if key not in ret:
            ret[key] = []

        ret[key].extend(value)
        ret[key] = sorted(ret[key], key=lambda entry: (entry["HostIp"], entry["HostPort"]))

    #print(ret)
    return ret


def _ports_match(desired: t.Mapping[int, t.Union[int, str]], ports: t.Mapping[str, t.Mapping[str, str]]):
    expected = my_format_to_ports_attr_format(desired)
    sorted_ports = dict()
    # TODO this breaks if the container EXPOSEs a port but I don't bind it
    for k, v in ports.items():
        if v:
            sorted_ports[k] = sorted(v, key=lambda entry: (entry["HostIp"], entry["HostPort"]))
    return expected == sorted_ports


def _envs_match(desired: DockerContainer, found: Container, desired_env):
    found_env_list: t.List[str] = found.attrs['Config']['Env']
    found_env = {l[0]: l[1] for l in (l.split("=", 1) for l in found_env_list) if l[0] in desired.env}
    return found_env == desired_env


def _volumes_match(desired: DockerContainer, found: Container):
    found_mounts_list: t.List[t.Dict[str, t.Any]] = found.attrs['Mounts']
    found_mounts_volume = {m['Name']: {"bind": m['Destination'], "mode": m["Mode"]} for m in found_mounts_list if m['Type'] == 'volume'}
    found_mounts_bind = {m['Source']: {"bind": m['Destination'], "mode": m["Mode"]} for m in found_mounts_list if m['Type'] == 'bind'}
    found_mounts = {}
    for k, v in found_mounts_volume.items():
        if k in desired.volumes:
            found_mounts[k] = v

    for k, v in found_mounts_bind.items():
        if k in desired.volumes:
            found_mounts[k] = v

    return found_mounts == build_volumes_dict(desired.volumes)


def _commands_match(desired: DockerContainer, found: Container):
    if desired.command:
        return desired.command.split(" ") == found.attrs["Config"]["Cmd"]
    else:
        return found.image.attrs['Config']['Cmd'] == found.attrs["Config"]["Cmd"]


def is_acceptable_container(c: Container, img_by_digest: Image, desired: DockerContainer, desired_env):
    return c.image.id == img_by_digest.id and \
           c.status in ["running"] and \
           _ports_match(desired.ports, c.ports) and \
           _envs_match(desired, c, desired_env) and \
           _volumes_match(desired, c) and \
           _commands_match(desired, c)





import datetime
import json
import typing as t

import requests
from requests import Response

DOCKER_DIGEST_HEADER = "Docker-Content-Digest"


class Registry:
    host: str
    proto: str

    def __init__(self, host, proto="http"):
        self.host = host
        self.proto = proto

    def list_repositories(self) -> t.List[str]:
        resp: Response = self._check_call(f"{self.proto}://{self.host}:5000/v2/_catalog")
        repos = resp.json()["repositories"]
        return repos

    def list_tags(self, repo: str) -> t.List[str]:
        resp: Response = self._check_call(f"{self.proto}://{self.host}:5000/v2/{repo}/tags/list")
        return resp.json()["tags"]

    def get_digest(self, repo: str, tag: str) -> str:
        resp: Response = self._check_call(f"{self.proto}://{self.host}:5000/v2/{repo}/manifests/{tag}", requests.get,
                                          headers={"Accept": "application/vnd.docker.distribution.manifest.v2+json"})
        digest = resp.headers[DOCKER_DIGEST_HEADER]
        return digest

    def get_create_date(self, repo: str, tag: str) -> datetime.datetime:
        resp: Response = self._check_call(f"{self.proto}://{self.host}:5000/v2/{repo}/manifests/{tag}", requests.get,
                                          headers={"Accept": "application/vnd.docker.distribution.manifest.v1+json"})
        raw_dt = json.loads(resp.json()["history"][0]["v1Compatibility"])["created"]
        return datetime.datetime.strptime(raw_dt[:len("yyyy-mm-ddThh:mm:ss.ffffff")], "%Y-%m-%dT%H:%M:%S.%f")

    def delete_image(self, repo: str, digest: str):
        self._check_call(f"{self.proto}://{self.host}:5000/v2/{repo}/manifests/{digest}", requests.delete)

    def get_contents(self) -> t.Dict[str, t.List['DockerImage']]:
        contents = {}
        for repo in self.list_repositories():
            tags = self.list_tags(repo)
            images = []
            for tag in tags:
                images.append(DockerImage(repo, tag, self.get_digest(repo, tag), self.get_create_date(repo, tag)))
            contents[repo] = images

        return contents

    def _check_call(self, url, fn: t.Callable = requests.get, headers=None):
        resp: Response = fn(url, headers=headers)
        if not resp.ok:
            raise ValueError(f"Request to {url} had code: {resp.status_code}: {resp.text}")
        return resp


class DockerImage(t.NamedTuple):
    repo: str
    tag: str
    digest: str
    created: datetime.datetime


import typing as t

import yaml


def loads(ident: str, necronomicon: str) -> 'Necronomicon':
    return load(ident, yaml.safe_load(necronomicon))


def loadfile(ident: str, path: str) -> 'Necronomicon':
    with open(path) as nin:
        return loads(ident, nin.read())


def load(ident: str, parsed_necronomicon) -> 'Necronomicon':
    if parsed_necronomicon is None:
        return Necronomicon(ident, NeededTunnelsSection([]), DockerSection([]), CronSection([]), FileSection([]))

    if 'files' in parsed_necronomicon:
        fs = FileSection([File(
            e['src'],
            e['dest'],
            e['hupcmd'] if 'hupcmd' in e else None,
            e['root'] if 'root' in e else False,
        ) for e in parsed_necronomicon['files']])
    else:
        fs = FileSection([])

    if 'docker' in parsed_necronomicon:
        ds = DockerSection([DockerContainer(
            e['image'],
            e['name'],
            e['fingerprint'],
            e['volumes'] if 'volumes' in e else {},
            {str(k): str(v) for k, v in (e['ports'] if 'ports' in e else {}).items()},
            e['env'] if 'env' in e else {},
            e['command'] if 'command' in e else None,
        ) for e in parsed_necronomicon['docker']])
    else:
        ds = DockerSection([])

    if 'cron' in parsed_necronomicon:
        cs = CronSection([CronJob(e['expr'], e['command'], e['user'] if 'user' in e else 'root') for e in parsed_necronomicon['cron']])
    else:
        cs = CronSection([])

    if 'needs_tunnels' in parsed_necronomicon:
        tunnels = NeededTunnelsSection([
            NeededTunnel(
                tun['host'],
                int(tun['target_port']),
                int(tun['local_port']),
            ) for tun in parsed_necronomicon['needs_tunnels']
        ])
    else:
        tunnels = NeededTunnelsSection([])

    return Necronomicon(ident, tunnels, ds, cs, fs)


class Necronomicon(t.NamedTuple):
    ident: str
    tunnels: 'NeededTunnelsSection'
    docker: 'DockerSection'
    cron: 'CronSection'
    files: 'FileSection'


class DockerSection(t.NamedTuple):
    containers: t.List['DockerContainer']


class DockerContainer(t.NamedTuple):
    image: str
    name: str
    fingerprint: str
    volumes: t.Mapping[str, str]
    ports: t.Mapping[str, str]
    env: t.Mapping[str, str]
    command: t.Optional[str]


class CronSection(t.NamedTuple):
    crons: t.List['CronJob']


class CronJob(t.NamedTuple):
    expr: str
    command: str
    user: str


class FileSection(t.NamedTuple):
    files: t.List['File']


class File(t.NamedTuple):
    src: str
    dest: str
    hupcmd: str
    root: bool


class NeededTunnel(t.NamedTuple):
    host: str
    target_port: int
    local_port: int


class NeededTunnelsSection(t.NamedTuple):
    tunnels: t.List[NeededTunnel]


import unittest
import yaml

from res import get_resource
from yog.host.manage_docker_utils import build_volumes_dict
from yog.host.necronomicon import load
from yog.host.manage import load_necronomicons_for_host
from tempfile import mkdtemp
from shutil import rmtree
from os.path import join, dirname
from os import makedirs


class TestNecronomicon(unittest.TestCase):
    def test(self):
        n = load("test1", yaml.safe_load(get_resource("sample_necronomicon.yml")))
        self.assertEqual("mail-relay-postfix", n.docker.containers[0].image)
        self.assertEqual(2, len(n.docker.containers))
        self.assertEqual(1, len(n.cron.crons))
        self.assertEqual("/var/lib/memes", n.docker.containers[1].volumes['memebox'])
        self.assertEqual("test.txt", n.files.files[0].src)

        n = load("test2", yaml.safe_load(get_resource("sample_needs_tunnel_back.yml")))
        self.assertEqual("test.josh.cafe", n.tunnels.tunnels[0].host)
        self.assertEqual(5000, n.tunnels.tunnels[0].target_port)
        self.assertEqual(5000, n.tunnels.tunnels[0].local_port)

    def test_load_necronomicons(self):
        root_dir = mkdtemp()
        try:
            host_file_path = "domains/cafe/josh/clust/ofn/myhost.yml"
            clust_file_path = "domains/cafe/josh/clust.yml"
            hostname = "myhost.ofn.clust.josh.cafe"

            makedirs(dirname(join(root_dir, host_file_path)), exist_ok=True)

            with open(join(root_dir, clust_file_path), "w") as out:
                out.write(get_resource("sample_site_necronomicon.yml").decode("utf-8"))
            with open(join(root_dir, host_file_path), "w") as out:
                out.write(get_resource("sample_necronomicon.yml").decode("utf-8"))

            necronomicons = load_necronomicons_for_host(hostname, root_dir)
            self.assertEqual(2, len(necronomicons))
            self.assertEqual("test2.txt", necronomicons[0].files.files[0].src)
            self.assertEqual("test.txt", necronomicons[1].files.files[0].src)
        finally:
            rmtree(root_dir)

    def test_build_volumes_dist(self):
        n = load("test1", yaml.safe_load(get_resource("sample_necronomicon.yml")))
        self.assertEqual("/tmp/test+ro", n.docker.containers[1].volumes['/tmp/test'])
        d = build_volumes_dict(n.docker.containers[1].volumes)
        self.assertEqual("ro", d["/tmp/test"]["mode"])
        self.assertEqual("rw", d["memebox"]["mode"])


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yog', 'yog.host', 'yog.repo', 'yog.res']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'docker>=5.0.3,<6.0.0', 'paramiko>=2.8.0,<3.0.0']

entry_points = \
{'console_scripts': ['yog = yog.host.main:main',
                     'yog-repo = yog.repo.main:main']}

setup_kwargs = {
    'name': 'yog',
    'version': '1.3.5',
    'description': 'The Gate and Key',
    'long_description': '# Yog\n\n## Overview\n\nAn opinionated docker-and-ssh-centric declarative system management tool.\n\n`pip3 install yog`\n\nFeatures:\n* Like puppet or ansible but a lot smaller and focused on docker, files, and cron\n* agentless - runs entirely on top of ssh\n* entirely defers auth(z/n) to ssh and the remote system\'s user permissions\n\nCommand summary:\n\n* `yog`: Applies configurations to hosts. e.g. `yog myhost.mytld` applies the config from `./domains/mytld/myhost.yml`.\n* `yog-repo`: Manages a docker repository. `yog-repo push` uses the contents of `./yog-repo.conf` to build an image and push it to the configured registry with the configured name and tag.\n\nExample run:\n\n[usage.webm](https://user-images.githubusercontent.com/1287152/209723654-e78b5283-60b5-4894-b5a1-3d2d71bfcc45.webm)\n\n\n## Setup\n\n1. Configure docker to listen on localhost\'s port 4243 (which is the default). See below.\n2. Use `ssh-copy-id` to copy your ssh key to all the servers you wish to manage. You can look into ssh certificates if you want a more general ssh PKI solution.\n3. Configure the target system to allow you to use sudo without a password. [2]\n4. That\'s it. You now serve the nameless mist. Do you hear whippoorwills?\n\n### Docker port listening setup (step 1)\n\n```bash\nssh myhost\nsudo systemctl edit docker\n```\n\nAnd add `-H tcp://127.0.0.1:4243` to the command. So for me, that file looks like:\n\n```text\n[Service]\nExecStart=\nExecStart=/usr/bin/dockerd -H fd:// -H tcp://127.0.0.1:4243\n```\n\nYog will apply files before docker containers, so you might also want to add a yog file entry like so:\n\n```yaml\nfiles:\n  - src: docker-override.conf\n    dest: /etc/systemd/system/docker.service.d/override.conf\n    root: yes\n    hupcmd: sudo systemctl restart docker\n```\n\n## Usage\n\nYog uses YML files that it calls "necronomicons" for configuration of hosts. It\'s organized hierarchically so\nthat a necronomicon for "mytld" will be applied to all hosts under that TLD.\n\nLet\'s learn by example:\n\nSuppose we have a folder, that can be named whatever we want, at `~/projects/my-config`. This is the root of a \ngit repo, and is also the root of our yog configuration. Make this your current working dir, or pass `--root-dir`.\n\n`$ cd ~/projects/my-config`\n\n```text\n.\n├── domains\n│      ├── com\n│      │      └── example\n│      │          └── myhost.yml\n│      └── com.yml\n└── files\n    ├── example.txt\n    ├── hello_world.html\n    └── helloworld-nginx.conf\n\n4 directories, 5 files\n```\n\nFiles that can be sent to hosts are stored under `files`.\n\nHost configurations - necronomicons - are stored under `domains`.\n\nIf we want to apply `myhost.yml`, we run:\n\n`yog myhost.example.com`\n\nExample output:\n```text\n$ yog myhost.example.com\n[2022-12-26 11:01:52,514] [INFO]: [myhost.example.com]\n[2022-12-26 11:01:59,121] [INFO]: [myhost.example.com][files]: OK [hello_world.html]\n[2022-12-26 11:01:59,274] [INFO]: [myhost.example.com][files]: OK [helloworld-nginx.conf]\n[2022-12-26 11:02:07,117] [INFO]: [myhost.example.com][docker]: OK registry@sha256:8be26f81ffea54106bae012c6f349df70f4d5e7e2ec01b143c46e2c03b9e551d\n```\n\n### Necronomicon format\n\nLet\'s look at a necronomicon.\n\n```yml\nfiles:\n  - src: hello_world.html\n    dest: /srv/hello_world/hello_world.html\n    root: yes\n  - src: helloworld-nginx.conf\n    dest: /etc/nginx/conf.d/helloworld.conf\n    root: yes\n    hupcmd: sudo systemctl restart nginx\n\n\ndocker:\n  - image: registry\n    name: my-registry\n    fingerprint: sha256:8be26f81ffea54106bae012c6f349df70f4d5e7e2ec01b143c46e2c03b9e551d\n    volumes:\n      images: /var/lib/registry\n    ports:\n      5000: 5000\n    env:\n      REGISTRY_STORAGE_DELETE_ENABLED: true\n```\n\n#### Files\n\nFiles are checked for equality via hash-comparison. I\'ve found this a useful way to manage:\n\n* cron files in /etc/cron.d/\n* Root certificates to put in the system trust store[1]\n* random config files\n\nAttributes:\n\n* `src`: the source. This is a _relative_ path rooted at the `files` directory in the hierarchy. You can use intermediate dirs.\n* `dest`: the destination filepath on the managed host. This is an absolute path.\n* `root`: whether to `sudo` to root for the file put. This mainly picks who owns the file + can access files, but this might have other useful properties for your use case. If set to `no`, the put operation is run as your ssh user. \n* `hupcmd`: a command to run after the file is placed. A common thing in ye olde days was to send SIGHUP to a process which would handle it by reloading the config. Commonly nowadays you might be using `hupcmd: sudo systemctl reload nginx`\n\n\n#### Docker containers\n\nDocker containers are compared on all specified attributes and won\'t unnecessarily restart containers. \n\nAttributes: \n\n* `image`: the docker repository name. e.g. `itzg/minecraft-server` or `dockerrepo.local:5000/mything`\n* `name`: the container name.\n* `fingerprint`: sha digest of the desired version. Tags are bad news bears so we don\'t support them. This is called fingerprint instead of digest because I didn\'t know they were called digests when I first coded this and then never changed it once I did.\n* `volumes`: volumes to attach. see below.\n* `ports`: ports to open. see below.\n* `env`: environment variables to set.\n\n##### Volumes\n\nFor volumes, the key is the volume name and the value is the mount point.\n\nFor bind mounts, the key is the host path and the value is the container path.\n\n##### Ports\n\nThe key is the host addr/port, and the value is the dest container port. Examples:\n\n```yml\n192.168.1.103:53/tcp: 53/tcp\n192.168.1.103:53/udp: 53/udp\n127.0.0.1:53/tcp: 53/tcp\n127.0.0.1:53/udp: 53/udp\n33200: 33200\n8080: 3000 # host port 8080 maps to container port 3000\n```\n\n## Footnotes\n\n[1] This is one of those things where I feel like you probably shouldn\'t manage root certs like this but I have yet to regret it? It\'s not a cryptographic secret, so.\n\n[2] Also something that people say you probably shouldn\'t do but I\'ve yet to regret. If your user is in the docker group it\'s basically root anyway from a threat modeling perspective.\n\n\n## `yog-repo`\n\nYog also includes a tool for pushing images to your local docker registry. I haven\'t documented it yet, apologies. \n',
    'author': 'Josh Hertlein',
    'author_email': 'jmhertlein@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
